/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is ForumZilla 0.1.1.
 *
 * The Initial Developer of the Original Code is Zapogee, Inc.
 * Portions created by Zapogee, Inc. are Copyright (C) 2001
 * Zapogee, Inc.  All Rights Reserved.
 *
 * Contributor(s):
 *   Myk Melez, <myk@melez.com>, original author
 */



/* constructor */

function Feed(resource) {
    this.resource = resource.QueryInterface(Components.interfaces.nsIRDFResource);

    this.description = null;
    this.author = null;
  
    this.request = null;
  
    return this;
}



/* getters and setters for properties stored in the database */

Feed.prototype.updateField = function(field, oldValue, newValue) {
    var ds = feedService.subscriptionsDS;
    if (oldValue)
        ds.Change(this.resource, field, oldValue, newValue);
    else
        ds.Assert(this.resource, field, newValue, true);
}

Feed.prototype.__defineGetter__('url',
function() {
    var ds = feedService.subscriptionsDS;
    var url = ds.GetTarget(this.resource, DC_IDENTIFIER, true);
    if (url)
      url = url.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
    else
      url = this.resource.Value;
    return url;
});

Feed.prototype.__defineSetter__('url',
function(newValue) {
    var ds = feedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, DC_IDENTIFIER, true);
    newValue = rdf.GetLiteral(newValue);
    this.updateField(DC_IDENTIFIER, oldValue, newValue);
    // XXX Update feeditems data source with new URL for references to this feed.
});

Feed.prototype.__defineGetter__('title',
function() {
    var ds = feedService.subscriptionsDS;
    var title = ds.GetTarget(this.resource, DC_TITLE, true);
    if (title)
      title = title.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
    return title;
});

Feed.prototype.__defineSetter__('title',
function(newValue) {
    var ds = feedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, DC_TITLE, true);
    newValue = rdf.GetLiteral(newValue);
    this.updateField(DC_TITLE, oldValue, newValue);
});

Feed.prototype.__defineGetter__('quickMode',
function() {
    var ds = feedService.subscriptionsDS;
    var quickMode = ds.GetTarget(this.resource, FZ_QUICKMODE, true);
    if (quickMode) {
        quickMode = quickMode.QueryInterface(Components.interfaces.nsIRDFLiteral);
        quickMode = quickMode.Value;
        quickMode = quickMode == "true" ? true : false;
    }
    return quickMode;
});

Feed.prototype.__defineSetter__('quickMode',
function(newValue) {
    var ds = feedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, FZ_QUICKMODE, true);
    newValue = newValue ? "true" : "false";
    newValue = rdf.GetLiteral(newValue);
    this.updateField(FZ_QUICKMODE, oldValue, newValue);
});

Feed.prototype.__defineGetter__('enabled',
function() {
    var ds = feedService.subscriptionsDS;
    var enabled = ds.GetTarget(this.resource, FZ_ENABLED, true);
    if (enabled) {
        enabled = enabled.QueryInterface(Components.interfaces.nsIRDFLiteral);
        enabled = enabled.Value;
        enabled = enabled == "true" ? true : false;
    }
    else
        enabled = true;
    return enabled;
});

Feed.prototype.__defineSetter__('enabled',
function(newValue) {
    var ds = feedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, FZ_ENABLED, true);
    newValue = newValue ? "true" : "false";
    newValue = rdf.GetLiteral(newValue);
    this.updateField(FZ_ENABLED, oldValue, newValue);
});

// XXX The getter returns an nsIMsgFolder, but the setter takes a URI. Hmmm.
Feed.prototype.__defineGetter__('folder',
function() {
    var ds = feedService.subscriptionsDS;
    var uri = ds.GetTarget(this.resource, FZ_FOLDER, true);

    if (uri) {
        uri = uri.QueryInterface(Components.interfaces.nsIRDFResource).Value;
        if (!GetMsgFolderFromUri(uri, true))
            feedService.createFolder(uri);
    }
    else {
        this.debug("no folder registered; checking for old school folder");
        uri = this.oldFolderURI;
        if (!uri) {
            this.debug("no old school folder; creating default folder");
            uri = feedService.createDefaultFolder(this.title);
        }
        this.debug("registering folder " + uri);
        ds.Assert(this.resource, FZ_FOLDER, rdf.GetResource(uri), true);
    }
    return GetMsgFolderFromUri(uri, true);
});

Feed.prototype.__defineSetter__('folder',
function(newValue) {
    var ds = feedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, FZ_FOLDER, true);
    newValue = rdf.GetResource(newValue);
    this.updateField(FZ_FOLDER, oldValue, newValue);
});

Feed.prototype.__defineGetter__('lastPolled',
function() {
    var ds = feedService.subscriptionsDS;
    var lastPolled = ds.GetTarget(this.resource, FZ_LASTPOLLED, true);
    if (lastPolled)
      lastPolled = new Date(lastPolled.QueryInterface(Components.interfaces.nsIRDFLiteral).Value);
    return lastPolled;
});

Feed.prototype.__defineSetter__('lastPolled',
function(newValue) {
    var ds = feedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, FZ_LASTPOLLED, true);
    newValue = newValue.toUTCString();
    newValue = rdf.GetLiteral(newValue);
    this.updateField(FZ_LASTPOLLED, oldValue, newValue);
});

Feed.prototype.__defineGetter__('interval',
function() {
    var ds = feedService.subscriptionsDS;
    var interval = ds.GetTarget(this.resource, FZ_INTERVAL, true);
    if (interval)
      interval = interval.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
    else
      interval = feedService.pollingInterval;
    return interval;
});

Feed.prototype.__defineSetter__('interval',
function(newValue) {
    var ds = feedService.subscriptionsDS;
    var oldValue = ds.GetTarget(this.resource, FZ_INTERVAL, true);
    newValue = rdf.GetLiteral(newValue);
    this.updateField(FZ_INTERVAL, oldValue, newValue);
});



/* getters and setters for generated properties */

// The name of the feed folder back when they had to be named after their feeds.
// XXX This can go away once enough users have upgraded from 0.5.0-.
Feed.prototype.__defineGetter__('oldFolderName',
function() {
    var name = this.title || this.description || this.url;
    if (!name)
        throw("couldn't compute feed name, as feed has no title, description, or URL.");
  
    // Make sure the feed name doesn't have any line breaks, since we're going
    // to use it as the name of the folder in the filesystem.  This may not
    // be necessary, since Mozilla's mail code seems to handle other forbidden
    // characters in filenames and can probably handle these as well.
    name = name.replace(/[\n\r\t]+/g, " ");
  
    // Make sure the feed doesn't end in a period to work around bug 117840.
    name = name.replace(/\.+$/, "");
  
    return name;
});

// The URI of the feed folder back when they had to be named after their feeds.
// XXX This can go away once enough users have upgraded from 0.5.0-.
Feed.prototype.__defineGetter__('oldFolderURI',
function() {
    var uri;
    try {
        // Try to get the folder by looking up the old pre-0.5.1 "News & Blogs"
        // account and seeing if there's a folder in there named after the feed.
        var accountManager =
            Components
                .classes["@mozilla.org/messenger/account-manager;1"]
                    .getService();
        accountManager = accountManager.QueryInterface(Components.interfaces.nsIMsgAccountManager);
        var accounts = accountManager.accounts.QueryInterface(Components.interfaces.nsICollection);
        var account, oldAccount;
        for ( var i=0 ; i<accounts.Count() ; i++ ) {
          account = accounts.GetElementAt(i);
          account = account.QueryInterface(Components.interfaces.nsIMsgAccount);
          if (account.incomingServer && account.incomingServer.prettyName == "News & Blogs")
            oldAccount = account;
            continue;
        }
        if (oldAccount
            && oldAccount.incomingServer
            && oldAccount.incomingServer.rootMsgFolder
            && oldAccount.incomingServer.rootMsgFolder.containsChildNamed(this.oldFolderName))
        {
            this.debug("old school folder for this feed; using it");
            uri = 
              oldAccount
                .incomingServer
                  .rootMsgFolder
                    .getChildNamed(this.oldFolderName)
                      .QueryInterface(Components.interfaces.nsIMsgFolder)
                        .URI;
        }
    }
    catch(e) {
        this.debug("error getting old school folder: " + e);
    }
    return uri;
});


Feed.prototype.__defineGetter__('charset',
function() {
    if (!this.request || !this.request.responseXML)
        return null;
   return this.request.responseXML.characterSet;
});



/* methods */
  
Feed.prototype.download = function(async, parseItems) {
    // Mark the feed as polled so we don't try to download again
    // until the polling interval has passed.
    this.lastPolled = new Date();

    // Whether or not to download the feed asynchronously.
    async = async == null ? true : async ? true : false;
  
    // Whether or not to parse items when downloading and parsing the feed.
    // Defaults to true, but setting to false is useful for obtaining
    // just the title of the feed when the user subscribes to it.
    this.parseItems = parseItems == null ? true : parseItems ? true : false;
  
    this.request = new XMLHttpRequest();
    this.request.open("GET", this.url, async);
    this.request.overrideMimeType("text/xml");
    if (async) {
        this.request.onload = Feed.onDownloaded;
        this.request.onerror = Feed.onDownloadError;
    }
    this.request.send(null);
    if (!async) {
        this.parse();
    }
}

Feed.onDownloaded = function(event) {
    var request = event.target;
    var url = request.channel.originalURI.spec;
    debug(url + " downloaded");
    var feed = feedService.getFeed(url);
    if (!feed)
        throw(url + " error after downloading; couldn't retrieve feed");
    feed.parse();
}

Feed.onDownloadError = function(event) {
    var request = event.target;
    var url = request.channel.originalURI.spec;
    var error = (request.status ? request.status : "") + " " +
                (request.statusText ? request.statusText : "")
    throw(url + " download failed: " + error);
}

Feed.prototype.parse = function() {
    // Figures out what description language (RSS, Atom) and version this feed
    // is using and calls a language/version-specific feed parser.
  
    //var start = new Date();
    //this.debug("parsing feed " + this.url + " at " + start);
  
    if (!this.request.responseText) {
        throw("error parsing feed " + this.url + ": no data");
        return;
    }
    else if (this.request.responseText.search(/="http:\/\/purl\.org\/rss\/1\.0\/"/) != -1) {
        this.debug("is an RSS 1.x (RDF-based) feed");
        this.parseAsRSS1();
    }
    else if (this.request.responseText.search(/="http:\/\/purl.org\/atom\/ns#"/) != -1) {
        this.debug("is an Atom feed");
        this.parseAsAtom();
    }
    else if (this.request.responseText.search(/"http:\/\/my\.netscape\.com\/rdf\/simple\/0\.9\/"/) != -1)
    {
        // RSS 0.9x is forward compatible with RSS 2.0, so use the RSS2 parser to handle it.
        this.debug("is an 0.9x feed");
        this.parseAsRSS2();
    }
    // XXX Explicitly check for RSS 2.0 instead of letting it be handled by the
    // default behavior (who knows, we may change the default at some point).
    else {
        // We don't know what kind of feed this is; let's pretend it's RSS 0.9x
        // and hope things work out for the best.  In theory even RSS 1.0 feeds
        // could be parsed by the 0.9x parser if the RSS namespace was the default.
        this.debug("is of unknown format; assuming an RSS 0.9x feed");
        this.parseAsRSS2();
    }
  
    //var end = new Date();
    //this.debug("done parsing feed " + this.url + " at " + end);
    //var cost = end - start;
    //gTotalCost += cost;
    //this.debug("cost for " + this.title + ": " + start + " to " + end + " = " + cost);
}

Feed.prototype.parseAsRSS2 = function() {
    if (!this.request.responseXML || !(this.request.responseXML instanceof XMLDocument))
        throw("error parsing RSS 2.0 feed " + this.url + ": data not parsed into XMLDocument object");

    // Get the first channel (assuming there is only one per RSS File).
    var channel = this.request.responseXML.getElementsByTagName("channel")[0];
    // If we can't find a channel element, look for stuff in the file as a whole.
    if (!channel)
        channel = this.request.responseXML;

    if (!this.title)
        this.title = getNodeValue(channel.getElementsByTagName("title")[0]);

    this.description = getNodeValue(channel.getElementsByTagName("description")[0]);

    if (!this.parseItems)
        return;

    this.invalidateItems();
  
    var itemNodes = this.request.responseXML.getElementsByTagName("item");
    for ( var i=0 ; i<itemNodes.length ; i++ ) {
        var itemNode = itemNodes[i];
        var item = new FeedItem();
        item.feed = this;
    
        var link = getNodeValue(itemNode.getElementsByTagName("link")[0]);
    
        var guidNode = itemNode.getElementsByTagName("guid")[0];
        if (guidNode) {
            var guid = getNodeValue(guidNode);
            var isPermaLink =
                guidNode.getAttribute('isPermaLink') == 'false' ? false : true;
        }
    
        item.url = (guid && isPermaLink) ? guid : link ? link : null;
        item.id = guid;
        item.description = getNodeValue(itemNode.getElementsByTagName("description")[0]);
        item.title = getNodeValue(itemNode.getElementsByTagName("title")[0])
                     || (item.description ? item.description.substr(0, 150) : null)
                     || item.title;
        item.author = getNodeValue(itemNode.getElementsByTagName("author")[0]
                                   || itemNode.getElementsByTagName("creator")[0]
                                   || channel.getElementsByTagName("creator")[0])
                      || this.title
                      || item.author;
        item.date = getNodeValue(itemNode.getElementsByTagName("pubDate")[0]
                                 || itemNode.getElementsByTagName("date")[0])
                    || item.date;
        var content = getNodeValue(itemNode.getElementsByTagNameNS(RSS_CONTENT_NS, "encoded")[0]);
        if (content)
            item.content = content;

        item.valid = true;
        if (!item.stored)
            feedService.queueItem(item);
    }

    this.removeInvalidItems();
}

Feed.prototype.parseAsRSS1 = function() {
    // RSS 1.0 is valid RDF, so use the RDF parser/service to extract data.
  
    // Create a new RDF data source and parse the feed into it.
    var ds = Components
               .classes["@mozilla.org/rdf/datasource;1?name=in-memory-datasource"]
                 .createInstance(Components.interfaces.nsIRDFDataSource);
    rdfparser.parseString(ds, this.request.channel.URI, this.request.responseText);
  
    // Get information about the feed as a whole.
    var channel = ds.GetSource(RDF_TYPE, RSS_CHANNEL, true);
    if (!this.title)
        this.title = getRDFTargetValue(ds, channel, RSS_TITLE);
    this.description = getRDFTargetValue(ds, channel, RSS_DESCRIPTION);

    if (!this.parseItems)
        return;
  
    this.invalidateItems();
  
    var items = ds.GetTarget(channel, RSS_ITEMS, true);
    try {
        if (!items)
            throw("no items container");
        //items = items.QueryInterface(Components.interfaces.nsIRDFContainer);
        items = rdfcontainer.MakeSeq(ds, items);
        items = items.GetElements();
        if (!items.hasMoreElements())
            throw("items container empty");
    }
    catch(e) {
        // For some reason we couldn't get a list of items from the items sequence,
        // (perhaps a buggy feed) so construct a list from all RSS item resources.
        items = ds.GetSources(RDF_TYPE, RSS_ITEM, true);
    }
  
    while (items.hasMoreElements()) {
        var itemResource = items.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
        var item = new FeedItem();
        item.feed = this;
    
        // Prefer the value of the link tag to the item URI since the URI could be
        // a relative URN.
        var uri = itemResource.Value;
        var link = getRDFTargetValue(ds, itemResource, RSS_LINK);
    
        item.url = link || uri;
        item.id = item.url;
        item.description = getRDFTargetValue(ds, itemResource, RSS_DESCRIPTION);
        item.title = getRDFTargetValue(ds, itemResource, RSS_TITLE)
                     || getRDFTargetValue(ds, itemResource, DC_SUBJECT)
                     || (item.description ? item.description.substr(0, 150) : null)
                     || item.title;
        item.author = getRDFTargetValue(ds, itemResource, DC_CREATOR)
                      || getRDFTargetValue(ds, channel, DC_CREATOR)
                      || this.title
                      || item.author;
        item.date = getRDFTargetValue(ds, itemResource, DC_DATE) || item.date;
        var content = getRDFTargetValue(ds, itemResource, RSS_CONTENT_ENCODED);
        if (content)
            item.content = content;
        item.valid = true;
        if (!item.stored)
            feedService.queueItem(item);
    }

    this.removeInvalidItems();
}

Feed.prototype.parseAsAtom = function() {
    if (!this.request.responseXML || !(this.request.responseXML instanceof XMLDocument))
      throw("error parsing Atom feed " + this.url + ": data not parsed into XMLDocument object");
  
    // Get the first channel (assuming there is only one per Atom File).
    var channel = this.request.responseXML.getElementsByTagName("feed")[0];
    if (!channel)
      throw("channel missing from Atom feed " + request.channel.name);

    if (!this.title)
        this.title = getNodeValue(channel.getElementsByTagName("title")[0]);
    this.description = getNodeValue(channel.getElementsByTagName("tagline")[0]);

    if (!this.parseItems)
      return;
  
    this.invalidateItems();
  
    var items = this.request.responseXML.getElementsByTagName("entry");
    for ( var i=0 ; i<items.length ; i++ ) {
        var itemNode = items[i];
        var item = new FeedItem();
        item.feed = this;
    
        var url;
        var links = itemNode.getElementsByTagName("link");
        for ( var j=0 ; j<links.length ; j++ ) {
            var alink = links[j];
            if (alink && alink.getAttribute('rel') && alink.getAttribute('rel') == 'alternate' && alink.getAttribute('href')) {
                url = alink.getAttribute('href');
                break;
            }
        }
    
        item.url = url;
        item.id = getNodeValue(itemNode.getElementsByTagName("id")[0]);
        item.description = getNodeValue(itemNode.getElementsByTagName("summary")[0]);
        item.title = getNodeValue(itemNode.getElementsByTagName("title")[0])
                     || (item.description ? item.description.substr(0, 150) : null)
                     || item.title;
    
        var author = itemNode.getElementsByTagName("author")[0]
                     || itemNode.getElementsByTagName("contributor")[0]
                     || channel.getElementsByTagName("author")[0];
        if (author) {
            var name = getNodeValue(author.getElementsByTagName("name")[0]);
            var email = getNodeValue(author.getElementsByTagName("email")[0]);
            if (name)
              author = name + (email ? " <" + email + ">" : "");
            else if (email)
              author = email;
            else
              author = null;
        }
        item.author = author || item.author || this.title;
    
        item.date = getNodeValue(itemNode.getElementsByTagName("modified")[0]
                                 || itemNode.getElementsByTagName("issued")[0]
                                 || itemNode.getElementsByTagName("created")[0])
                    || item.date;
    
        // XXX We should get the xml:base attribute from the content tag as well
        // and use it as the base HREF of the message.
        // XXX Atom feeds can have multiple content elements; we should differentiate
        // between them and pick the best one.
        // Some Atom feeds wrap the content in a CTYPE declaration; others use
        // a namespace to identify the tags as HTML; and a few are buggy and put
        // HTML tags in without declaring their namespace so they look like Atom.
        // We deal with the first two but not the third.
        var content;
        var contentNode = itemNode.getElementsByTagName("content")[0];
        if (contentNode) {
            content = "";
            for ( j=0 ; j<contentNode.childNodes.length ; j++ ) {
                var node = contentNode.childNodes.item(j);
                if (node.nodeType == node.CDATA_SECTION_NODE)
                    content += node.data;
                else
                    content += serializer.serializeToString(node);
                    //content += getNodeValue(node);
            }
            if (contentNode.getAttribute('mode') == "escaped") {
                content = content.replace(/&lt;/g, "<");
                content = content.replace(/&gt;/g, ">");
                content = content.replace(/&amp;/g, "&");
            }
            if (content == "")
                content = null;
        }
        item.content = content;
    
        item.valid = true;
        if (!item.stored)
            feedService.queueItem(item);
    }
  
    this.removeInvalidItems();
}

Feed.prototype.invalidateItems = function invalidateItems() {
    var ds = feedService.itemsDS;
    this.debug("invalidating items");
    var items = ds.GetSources(FZ_FEED, this.resource, true);
    var item;
    while (items.hasMoreElements()) {
        item = items.getNext();
        item = item.QueryInterface(Components.interfaces.nsIRDFResource);
        this.debug("invalidating " + item.Value);
        var valid = ds.GetTarget(item, FZ_VALID, true);
        if (valid)
            ds.Unassert(item, FZ_VALID, valid, true);
    }
}

Feed.prototype.removeInvalidItems = function() {
    var ds = feedService.itemsDS;
    this.debug("removing invalid items");
    var items = ds.GetSources(FZ_FEED, this.resource, true);
    var item;
    while (items.hasMoreElements()) {
        item = items.getNext();
        item = item.QueryInterface(Components.interfaces.nsIRDFResource);
        if (ds.HasAssertion(item, FZ_VALID, RDF_LITERAL_TRUE, true))
            continue;
        this.debug("removing " + item.Value);
        ds.Unassert(item, FZ_FEED, this.resource, true);
        if (ds.hasArcOut(item, FZ_FEED))
            this.debug(item.Value + " is from more than one feed; only the reference to this feed removed");
        else
            removeAssertions(ds, item);
    }
}

Feed.prototype.debug = function(txt) {
    debug(this.url + " " + txt);
}
